body {
    background-color: white;
    color: black;
    margin: 0;
    height: 100%;
    min-height: 100vh;
    display: flex;
}

.time {
    color: rgba(128, 128, 128, 0.7);
    display: none;
}
.sectionTitle {
    color: darkslateblue;
    font-weight: bold;
}
.successStatus {
    color: green;
    font-weight: bold;
}
.failedStatus {
    color: red;
    font-weight: bold;
}
.skippedStatus {
    color: red;
    font-weight: bold;
}
.skipped {
    color: darkcyan;
}
.exceptionTag {
    color: darkred;
}
.stderrTag {
    color: orange;
}

.hide {
    display: none;
}
.hide + label ~ div {
    display: none;
}
.hide + label {
    cursor: pointer;
    display: inline-block;
}
.hide:checked + label + div {
    display: block;
}

.hide + label + div + .nl {
    display: block;
}
.hide:checked + label + div + .nl::after {
    display: none;
}

.hide + label .sectionLbl::before {
    content: "[+] ";
}
.hide:checked + label .sectionLbl::before {
    content: "[-] ";
}

#dark-checkbox {
    display: none;
}

#time-checkbox {
    display: none;
}
pre {
    padding: 20px 20px 65px 20px;
    margin: 0;
    width: 100%;
}

#dark-checkbox:checked~pre {
    background-color: black;
    color: rgb(219, 198, 198);
}

#dark-checkbox:checked~pre .sectionTitle {
    color: #2b7cdf;
}

#dark-checkbox+label {
    position: fixed;
    right: 15px;
    bottom: 15px;
    width: 95px;
    height: 30px;
    border-radius: 20px;
    background-color: white;
    color: gray;
    border: gray 1px solid;
    font: 12px sans;
    cursor: pointer;
}

#dark-checkbox:checked+label {
    background-color: black;
    color: white;
}

#dark-checkbox+label::before {
    position: absolute;
    content: "";
    height: 22px;
    width: 22px;
    left: 4px;
    bottom: 4px;
    background-color: gray;
    transition: .3s;
    border-radius: 50%;
}

#dark-checkbox:checked+label::before {
    background-color: white;
    transform: translateX(65px);
}

#dark-checkbox+label::after {
    content: 'Light';
    display: block;
    position: absolute;
    transform: translate(-50%, -50%);
    top: 50%;
    left: 50%;
}

#dark-checkbox:checked+label::after {
    content: 'Dark';
}

#time-checkbox:checked+label~pre .time {
    display: inline-block;
}

#time-checkbox+label {
    position: fixed;
    right: 125px;
    bottom: 15px;
    width: 95px;
    height: 30px;
    border-radius: 20px;
    background-color: white;
    color: gray;
    border: gray 1px solid;
    font: 12px sans;
    cursor: pointer;
}

#dark-checkbox:checked~#time-checkbox+label {
    background-color: black;
    color: white;
}

#dark-checkbox:checked~#time-checkbox+label::before {
    background-color: white;
}

#time-checkbox:checked+label {
    background-color: rgb(24, 61, 16) !important;
    color: white;
}

#time-checkbox+label::before {
    position: absolute;
    content: "";
    height: 22px;
    width: 22px;
    left: 4px;
    bottom: 4px;
    background-color: gray;
    transition: .3s;
    border-radius: 50%;
}

#time-checkbox:checked+label::before {
    background-color: white;
    transform: translateX(65px);
}

#time-checkbox+label::after {
    content: 'Time';
    display: block;
    position: absolute;
    transform: translate(-50%, -50%);
    top: 50%;
    left: 50%;
}
